/* cc-kb-layout-dialog.c
 *
 * Copyright 2019 Bastien Nocera <hadess@hadess.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#include <glib/gi18n.h>
#define HANDY_USE_UNSTABLE_API
#include <handy.h>

#include "cc-kb-layout-dialog.h"
#include "list-box-helper.h"
#include "cc-keyboard-item.h"

struct _CcKbLayoutDialog
{
  GtkDialog       parent_instance;
  GtkLabel       *description_label;
  GtkListBox     *listbox;
  GtkScrolledWindow  *list_scrolled_window;
  GSettings      *input_source_settings;
  const InSrcModifier *modifier;
  GSList         *radio_group;
  CcKeyboardItem     *item;
  gboolean backward;
};

G_DEFINE_TYPE (CcKbLayoutDialog, cc_kb_layout_dialog, GTK_TYPE_DIALOG)

static const gchar *custom_css =
".xkb-option-button {"
"    padding: 12px"
"}";

static GtkRadioButton *
get_radio_button_default (CcKbLayoutDialog *self)
{
  GSList *l;

  for (l = self->radio_group; l != NULL; l = l->next)
    {
      if(g_object_get_data (l->data, "default-opt") != NULL)
        return l->data;
    }
  return NULL;
}

static GtkRadioButton *
get_radio_button_from_in_src_option (CcKbLayoutDialog *self)
{
  gchar *sw_option_1;
  gchar *sw_option_2;
  gchar *sw_option_3;
  gchar *sw_option_4;
  GSList *l;
  g_auto(GStrv) options = NULL;

  if(self->backward) {
    options = g_settings_get_strv (self->input_source_settings, "switch-input-source-backward");
  } else {
    options = g_settings_get_strv (self->input_source_settings, "switch-input-source");
  }

  if(options == NULL)
    return NULL;

  if(options[0] == NULL)
    return NULL;

  for (l = self->radio_group; l != NULL; l = l->next)
  {
    sw_option_1 = g_object_get_data (l->data, "switch-input-source-1");
    sw_option_2 = g_object_get_data (l->data, "switch-input-source-2");
    sw_option_3 = g_object_get_data (l->data, "switch-input-source-3");
    sw_option_4 = g_object_get_data (l->data, "switch-input-source-4");
    if (g_strcmp0 (sw_option_1, options[0]) == 0) {
      if(options[1] != NULL && sw_option_2 != NULL) {
        if (g_strcmp0 (sw_option_2, options[1]) == 0) {
          if(options[2] != NULL && sw_option_3 != NULL) {
            if (g_strcmp0 (sw_option_3, options[2]) == 0) {
              if(options[3] != NULL && sw_option_4 != NULL) {
                if (g_strcmp0 (sw_option_4, options[3]) == 0)
                  return l->data;
              }
            }
          }
        }
      }
    }
  }

  for (l = self->radio_group; l != NULL; l = l->next)
  {
    sw_option_1 = g_object_get_data (l->data, "switch-input-source-1");
    sw_option_2 = g_object_get_data (l->data, "switch-input-source-2");
    sw_option_3 = g_object_get_data (l->data, "switch-input-source-3");
    if (g_strcmp0 (sw_option_1, options[0]) == 0) {
      if(options[1] != NULL && sw_option_2 != NULL) {
        if (g_strcmp0 (sw_option_2, options[1]) == 0) {
          if(options[2] != NULL && sw_option_3 != NULL) {
            if (g_strcmp0 (sw_option_3, options[2]) == 0)
              return l->data;
            }
          }
       }
    }
  }

  for (l = self->radio_group; l != NULL; l = l->next)
  {
    sw_option_1 = g_object_get_data (l->data, "switch-input-source-1");
    sw_option_2 = g_object_get_data (l->data, "switch-input-source-2");
    if (g_strcmp0 (sw_option_1, options[0]) == 0) {
      if(options[1] != NULL && sw_option_2 != NULL) {
        if (g_strcmp0 (sw_option_2, options[1]) == 0)
          return l->data;
      }
    }
  }

  for (l = self->radio_group; l != NULL; l = l->next)
  {
    sw_option_1 = g_object_get_data (l->data, "switch-input-source-1");
    if (g_strcmp0 (sw_option_1, options[0]) == 0)
       return l->data;
  }

  return NULL;
}

static void
update_active_radio (CcKbLayoutDialog *self)
{
  g_auto(GStrv) options = NULL;
  const InSrcOpt *default_option;
  GtkRadioButton *rightalt_radio;
  GtkRadioButton *radio;

  if(self->item != NULL) {
    if (g_strcmp0 (cc_keyboard_item_get_key(self->item), "switch-input-source-backward") == 0) {
      gtk_label_set_markup (self->description_label, gettext (self->modifier->description_back));
      self->backward = TRUE;
    } else {
      gtk_label_set_markup (self->description_label, gettext (self->modifier->description));
      self->backward = FALSE;
    }
  }

  radio = get_radio_button_from_in_src_option (self);

  if (radio) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radio), TRUE);
    return;
  }

  rightalt_radio = get_radio_button_default(self);
  if(rightalt_radio != NULL) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (rightalt_radio), TRUE);
  }
}

static void
set_in_src_option (CcKbLayoutDialog *self,
               gchar               *switch_input_source_opt_1,
               gchar               *switch_input_source_opt_2,
               gchar               *switch_input_source_opt_3,
               gchar               *switch_input_source_opt_4)
{
  g_autoptr(GPtrArray) array_b = NULL;

  array_b = g_ptr_array_new ();
  if(switch_input_source_opt_1 != NULL) {
    g_ptr_array_add (array_b, switch_input_source_opt_1);

      if(switch_input_source_opt_2 != NULL)
        g_ptr_array_add (array_b, switch_input_source_opt_2);

      if(switch_input_source_opt_3 != NULL)
        g_ptr_array_add (array_b, switch_input_source_opt_3);

      if(switch_input_source_opt_4 != NULL)
        g_ptr_array_add (array_b, switch_input_source_opt_4);

      g_ptr_array_add (array_b, NULL);
      if(self->backward) {
        g_settings_set_strv (self->input_source_settings,
                      "switch-input-source-backward",
                      (const gchar * const *) array_b->pdata);
      } else {
                g_settings_set_strv (self->input_source_settings,
                      "switch-input-source",
                      (const gchar * const *) array_b->pdata);
      }
  }
}

static void
on_active_radio_changed_cb (CcKbLayoutDialog *self,
                            GtkRadioButton      *radio)
{
  gchar *switch_input_source_opt_1;
  gchar *switch_input_source_opt_2;
  gchar *switch_input_source_opt_3;
  gchar *switch_input_source_opt_4;

  if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (radio)))
    return;

  switch_input_source_opt_1 = (gchar *)g_object_get_data (G_OBJECT (radio), "switch-input-source-1");
  switch_input_source_opt_2 = (gchar *)g_object_get_data (G_OBJECT (radio), "switch-input-source-2");
  switch_input_source_opt_3 = (gchar *)g_object_get_data (G_OBJECT (radio), "switch-input-source-3");
  switch_input_source_opt_4 = (gchar *)g_object_get_data (G_OBJECT (radio), "switch-input-source-4");
  set_in_src_option (self, switch_input_source_opt_1, switch_input_source_opt_2, switch_input_source_opt_3, switch_input_source_opt_4);
}

static void
on_in_src_options_changed_cb (CcKbLayoutDialog *self)
{
  if (self->modifier == NULL)
    update_active_radio (self);
}

static gboolean
enable_switch_changed_cb (GtkSwitch *widget,
                          gboolean   state,
                          gpointer   user_data)
{
  CcKbLayoutDialog *self = user_data;
  gchar *switch_input_source_opt_1;
  gchar *switch_input_source_opt_2;
  gchar *switch_input_source_opt_3;
  gchar *switch_input_source_opt_4;
  GSList *l;

  gtk_widget_set_sensitive (GTK_WIDGET (self->listbox), state);

  if (state)
    {
      for (l = self->radio_group; l != NULL; l = l->next)
        {
          if (gtk_toggle_button_get_active (l->data))
            {
              switch_input_source_opt_1 = (gchar *)g_object_get_data (l->data, "switch-input-source-1");
              switch_input_source_opt_2 = (gchar *)g_object_get_data (l->data, "switch-input-source-2");
              switch_input_source_opt_3 = (gchar *)g_object_get_data (l->data, "switch-input-source-3");
              switch_input_source_opt_4 = (gchar *)g_object_get_data (l->data, "switch-input-source-4");
              set_in_src_option (self, switch_input_source_opt_1, switch_input_source_opt_2, switch_input_source_opt_3, switch_input_source_opt_4);
              break;
            }
        }
    }
  else
    {
      set_in_src_option (self, NULL, NULL, NULL, NULL);
    }

  return FALSE;
}

static void
cc_kb_layout_dialog_finalize (GObject *object)
{
  CcKbLayoutDialog *self = (CcKbLayoutDialog *)object;

  g_clear_object (&self->input_source_settings);

  G_OBJECT_CLASS (cc_kb_layout_dialog_parent_class)->finalize (object);
}

static void
cc_kb_layout_dialog_class_init (CcKbLayoutDialogClass *klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->finalize = cc_kb_layout_dialog_finalize;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/control-center/keyboard/cc-kb-layout-dialog.ui");

  gtk_widget_class_bind_template_child (widget_class, CcKbLayoutDialog, description_label);
  gtk_widget_class_bind_template_child (widget_class, CcKbLayoutDialog, listbox);
  gtk_widget_class_bind_template_child (widget_class, CcKbLayoutDialog, list_scrolled_window);
  gtk_widget_class_bind_template_callback (widget_class, enable_switch_changed_cb);
}

static void
add_radio_buttons (CcKbLayoutDialog *self)
{
  GtkWidget *row, *radio_button, *label, *last_button = NULL;
  InSrcOpt *options = self->modifier->options;
  int i;

  for (i = 0; options[i].label && options[i].switch_input_source_1; i++)
    {
      row = g_object_new (GTK_TYPE_LIST_BOX_ROW,
                          "visible", TRUE,
                          "selectable", FALSE,
                          NULL);
      gtk_container_add (GTK_CONTAINER (self->listbox), row);

      radio_button = g_object_new (GTK_TYPE_RADIO_BUTTON,
                                   "visible", TRUE,
                                   "can_focus", TRUE,
                                   "receives_default", FALSE,
                                   "draw_indicator", TRUE,
                                   NULL);
      label = g_object_new (GTK_TYPE_LABEL,
                            "visible", TRUE,
                            "margin_left", 6,
                            "label", g_dpgettext2 (NULL, "keyboard key", options[i].label),
                            NULL);
      gtk_container_add (GTK_CONTAINER (radio_button), label);
      gtk_style_context_add_class (gtk_widget_get_style_context (radio_button), "xkb-option-button");
      gtk_radio_button_join_group (GTK_RADIO_BUTTON (radio_button), GTK_RADIO_BUTTON (last_button));
      g_object_set_data (G_OBJECT (radio_button), "switch-input-source-1", options[i].switch_input_source_1);
      g_object_set_data (G_OBJECT (radio_button), "switch-input-source-2", options[i].switch_input_source_2);
      g_object_set_data (G_OBJECT (radio_button), "switch-input-source-3", options[i].switch_input_source_3);
      g_object_set_data (G_OBJECT (radio_button), "switch-input-source-4", options[i].switch_input_source_4);
      if(options[i].default_opt) {
        g_object_set_data (G_OBJECT (radio_button), "default-opt", "default");
      } else {
        g_object_set_data (G_OBJECT (radio_button), "default-opt", NULL);
      }
      g_signal_connect_object (radio_button, "toggled", (GCallback)on_active_radio_changed_cb, self, G_CONNECT_SWAPPED);
      gtk_container_add (GTK_CONTAINER (row), radio_button);

      last_button = radio_button;
    }

  self->radio_group = NULL;
  if (last_button != NULL)
    self->radio_group = gtk_radio_button_get_group (GTK_RADIO_BUTTON (last_button));
}

static void
cc_kb_layout_dialog_init (CcKbLayoutDialog *self)
{
  g_autoptr(GtkCssProvider) provider = NULL;

  gtk_widget_init_template (GTK_WIDGET (self));

  provider = gtk_css_provider_new ();
  gtk_css_provider_load_from_data (provider, custom_css, -1, NULL);

  gtk_style_context_add_provider_for_screen (gdk_screen_get_default (),
                                             GTK_STYLE_PROVIDER (provider),
                                             GTK_STYLE_PROVIDER_PRIORITY_APPLICATION + 1);

  self->modifier = NULL;

  gtk_list_box_set_header_func (self->listbox, cc_list_box_update_header_func, NULL, NULL);

  self->input_source_settings = g_settings_new ("org.gnome.desktop.wm.keybindings");
  g_signal_connect_object (self->input_source_settings,
                           "changed::switch-input-source-backward",
                           G_CALLBACK (on_in_src_options_changed_cb),
                           self, G_CONNECT_SWAPPED);
}

CcKbLayoutDialog *
cc_kb_layout_dialog_new (GSettings *input_settings,
                            const InSrcModifier *modifier)
{
  CcKbLayoutDialog *self;

  self = g_object_new (CC_TYPE_KB_LAYOUT_DIALOG,
                       "use-header-bar", TRUE,
                       NULL);
  self->input_source_settings = g_object_ref (input_settings);

  self->modifier = modifier;
  self->item = NULL;
  gtk_window_set_title (GTK_WINDOW (self), gettext (modifier->title));

  add_radio_buttons (self);
  update_active_radio (self);
  gtk_widget_set_sensitive (GTK_WIDGET (self->listbox), TRUE);

  return self;
}

void
cc_kb_layout_dialog_set_item (CcKbLayoutDialog *self,
                                      CcKeyboardItem           *item)
{
  self->item = item;
  update_active_radio (self);
}
